#
#  keep_best_hit_when_query_overlaps.pl
#
#  Version 0.1.1 (June 27, 2012)
#
#  Copyright (c) 2012 Kirill Kryukov
#
#  This software is provided 'as-is', without any express or implied
#  warranty. In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#

use strict;


my @number_to_alphanumeric_code = (
    '0','1','2','3','4','5','6','7','8','9',
    'A','B','C','D','E','F','G','H','I','J',
    'K','L','M','N','O','P','Q','R','S','T',
    'U','V','W','X','Y','Z','a','b','c','d',
    'e','f','g','h','i','j','k','l','m','n',
    'o','p','q','r','s','t','u','v','w','x',
    'y','z'
);

my @alphanumeric_code_to_number = (-1) x 256;

my @number_to_symbol_code = (
    '!','"','#','$','%','&',"'",'(',')','*',
    '+',',','-','.','/',':',';','<','=','>',
    '?','@','[','\\',']','^','_','`','{','|',
    '}','~'
);

my @symbol_code_to_number = (-1) x 256;

for (my $i=0; $i<62; $i++)
{
    $alphanumeric_code_to_number[ord($number_to_alphanumeric_code[$i])] = $i;
}
for (my $i=0; $i<32; $i++)
{
    $symbol_code_to_number[ord($number_to_symbol_code[$i])] = $i;
}



my ($q_id,$s_id) = ('','');
my @hits = ();
my ($n_processed,$n_kept) = (0,0);

while (<STDIN>)
{
    chomp;

    if (/^#/)
    {
        process();
        $q_id = $1;
        $s_id = '';
        @hits = ();
        print $_, "\n";
        next;
    }

    if (/^\s{2}/)
    {
        process();
        if (!/^\s{2}(\S+)$/) { die "Can't parse query index \"$_\"\n"; }
        $q_id = $1;
        $s_id = '';
        @hits = ();
        next;
    }

    if (/^\s/)
    {
        if (!/^\s(\S+)$/) { die "Can't parse subject index \"$_\"\n"; }
        $s_id = $1;
        next;
    }

    if ($q_id eq '' || $s_id eq '') { die "Hit with unknown query or subject: \"$_\"\n"; }

    if (!/^(\S)(\S+)\s(\S+)\s(\S*)\s(\S*)\s(\S+)\s(\S+)\s(\S*)\s(\S+)\s(\S+)(.*)$/) { die "Can't parse hit \"$_\"\n"; }
    my ($frame_code,$aln_length_code,$identities_code,$positives_code,$ngaps_code,$score_code,$e_code,$nparts_code,$qstart_code,$sstart_code,$qsgaps) = ($1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11);

    my $aln_length = decode_integer_base94($aln_length_code);
    my $qstart = decode_integer_base94($qstart_code);
    my ($qngaps,$qlen) = (0,-1);
    my @qint = ();

    if ($qsgaps eq '')
    {
        $qlen = $aln_length;
    }
    else
    {
        if ($qsgaps !~ /^\s(\S*)\s(\S*)$/) { die "Can't parse gap patterns, hit: \"$_\"\n"; }
        my ($q_gaps_code,$s_gaps_code) = ($1,$2);
        if ($q_gaps_code ne '') { parse_gap_pattern($q_gaps_code,\@qint,\$_); }
        for (my $i=1; $i<scalar(@qint); $i+=2) { $qngaps += $qint[$i]; }
        $qlen = $aln_length - $qngaps;
    }

    my $score = decode_scientific($score_code);
    my $e = decode_scientific($e_code);

    push @hits, {'sid' => $s_id, 'qs' => $qstart, 'qe' => $qstart + $qlen - 1, 'e' => $e, 'sc' => $score, 'h' => $_};
}
process();
print STDERR "Done, $n_kept hits kept out of $n_processed\n";



sub process
{
    if ($q_id eq '') { return; }

    my $n_in = scalar @hits;
    my $n_out = 0;

    my @ind = ();
    for (my $i=0; $i<$n_in; $i++) { $ind[$i] = $i; }
    @ind = sort { $hits[$a]->{'qs'} <=> $hits[$b]->{'qs'} } @ind;

    for (my $i=0; $i<$n_in-1; $i++)
    {
        if ($hits[$ind[$i]]->{'h'} eq '') { next; }
        for (my $j=$i+1; $j<$n_in and ($hits[$ind[$i]]->{'h'} ne '') and ($hits[$ind[$j]]->{'qs'} <= $hits[$ind[$i]]->{'qe'}); $j++)
        {
            if ($hits[$ind[$j]]->{'h'} eq '') { next; }
            my $rem = ( ($hits[$ind[$j]]->{'e'} < $hits[$ind[$i]]->{'e'}) or
                        ($hits[$ind[$j]]->{'e'} == $hits[$ind[$i]]->{'e'} and $hits[$ind[$j]]->{'sc'} > $hits[$ind[$i]]->{'sc'}) ) ? $i : $j;
            $hits[$ind[$rem]]->{'h'} = '';
        }
    }

    print '  ', $q_id, "\n";
    $s_id = '';
    for (my $i=0; $i<$n_in; $i++)
    {
        if ($hits[$i]->{'h'} eq '') { next; }
        if ($hits[$i]->{'sid'} ne $s_id)
        {
            $s_id = $hits[$i]->{'sid'};
            print ' ', $s_id, "\n";
        }
        #print $hits[$i]->{'qs'}, '..', $hits[$i]->{'qe'}, ' ', $hits[$i]->{'h'}, "\n";
        print $hits[$i]->{'h'}, "\n";
        $n_out++;
    }  

    $n_processed += $n_in;
    $n_kept += $n_out;
}



sub decode_integer_base94
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 94 + ord(substr($s,$i,1)) - 33;
    }
    return $n;
}



sub decode_integer_alphanumeric
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 62 + $alphanumeric_code_to_number[ord(substr($s,$i,1))];
    }
    return $n;
}



sub decode_integer_symbol
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 32 + $symbol_code_to_number[ord(substr($s,$i,1))];
    }
    return $n;
}



sub decode_float
{
    my ($code) = @_;
    my $di = index($code,'.');
    if ($di >= 0)
    {
        return decode_integer_alphanumeric(substr($code,0,$di)) . '.' . reverse(decode_integer_alphanumeric(substr($code,$di+1)));
    }
    else
    {
        return decode_integer_alphanumeric($code);
    }
}



sub decode_scientific
{
    my ($code) = @_;
    my $pi = index($code,'+');
    my $mi = index($code,'-');
    if ($pi >= 0 or $mi >= 0)
    {
        my $di = ($pi >= 0 and $mi >= 0) ? (($pi < $mi) ? $pi : $mi) : (($pi >= 0) ? $pi : $mi);
        return decode_float(substr($code,0,$di)) . 'e' . substr($code,$di,1) . sprintf("%03d",decode_integer_base94(substr($code,$di+1)));
    }
    else
    {
        return decode_float($code);
    }
}



sub parse_gap_pattern
{
    my ($code,$int,$hit) = @_;
    if ($alphanumeric_code_to_number[ord(substr($code,0,1))] < 0) { die "Can't parse gap pattern \"$code\" in hit \"$$hit\"\n"; }
    my $pe = -1;
    my $s = '';
    my $a = 1;
    my $l = length($code);
    for (my $i=0; $i<$l; $i++)
    {
        my $c = substr($code,$i,1);
        my $ca = ($alphanumeric_code_to_number[ord($c)] >= 0) ? 1 : 0;
        if ($ca == $a) { $s .= $c; next; }
        push @{$int}, ($a ? decode_integer_alphanumeric($s) : decode_integer_symbol($s));
        $s = $c;
        $a = $ca;
    }
    if ($a) { die "Last part of the gap pattern \"$code\" encodes non-gaps in hit \"$$hit\"\n"; }
    push @{$int}, decode_integer_symbol($s);
}
